(* {{{ COPYING *(

  CGI for a web site indexing free-creations around the web.
  Copyright (C) 2006, 2008  Florent Monnier

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation,
  Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

)* }}} *)
let start = Unix.gettimeofday() ;;
print_string "Content-Type: text/html\r\n\r\n" ;;

(* {{{ utils *)

let echo = print_string ;;
let echos = List.iter print_string ;;

let error msg = Printf.printf "<!-- Error:\n%s\n-->\n" msg ;;

let incoming_args = Cgi.parse_args() ;;

let datas_path = "/home/html/LO/FC/datas/" ;;

let script_name =
  try Unix.getenv "SCRIPT_NAME"
  with Not_found -> "./FC.cgi"
;;

let get_licence_url ~licence =
  match licence with
  | "CC:by-sa:2.5" -> "http://creativecommons.org/licenses/by-sa/2.5/"
  | "CC:by-sa:2.0" -> "http://creativecommons.org/licenses/by-sa/2.0/"
  | "CC:by:2.0" -> "http://creativecommons.org/licenses/by/2.0/"
  | "CC:by:2.5" -> "http://creativecommons.org/licenses/by/2.5/"
  | "GPL" -> "http://www.fsf.org/licensing/licenses/gpl.html"
  | "LAL" -> "http://artlibre.org/licence/lal/en/"
  | _ -> "./FC.cgi?bug=Licence_not_found"
;;

let md5 str = Digest.to_hex(Digest.string str) ;;

(* code copy-pasted from cgi.ml *)
let parse_sess_env assocs =
  let one_assoc s =
    try
      let i = String.index s '=' in
      String.sub s 0 i, 
      Cgi.decode (String.sub s (succ i) (String.length s - i - 1))
    with
      Not_found -> s,""
  in
  List.map one_assoc assocs
;;


(* {{{ DEBUG *(

let () =
  echo "<!-- Unix::ENV\n";
  Array.iter (Printf.printf "\t%s\n") (Unix.environment());
  echo "-->\n";
;;

let () =
  echo "<!-- incoming_args\n";
  List.iter (fun (k,v) -> Printf.printf " %s => '%s'\n" k v) incoming_args;
  echo "-->\n";
;;

)* }}} *)

(* }}} *)
(* {{{ Log User Domain *)

let log_domain() =
  let log_perms = 0o644 in
  let log_file = "./logs/logs" in
  (*
  Unix.chmod log_file log_perms;
  *)
  try
    let ip = Unix.getenv "REMOTE_ADDR" in
    let domain =
      let addr = Unix.inet_addr_of_string ip in
      let host = (Unix.gethostbyaddr addr) in
      (host.Unix.h_name)
    in
    let log_str = (Printf.sprintf "%s %s\n" ip domain) in
    let fd = Unix.openfile log_file [Unix.O_WRONLY; Unix.O_APPEND; Unix.O_CREAT] log_perms in
    let oc = Unix.out_channel_of_descr fd in
    output_string oc log_str;
    flush oc;
    Unix.close fd;
  with e ->
    error (Printexc.to_string e);
;;

(* }}} *)
(* {{{ session *)
(* If you wish, you can use the code related to the session management
   under LGPL licence, but be warn that it is unsafe and basic. *)

let make_sess_file ~sess_id ~user_agent ~ip =
  try
    let sess_file = ("/tmp/sess_" ^ sess_id)
    and perms = 0o644 in
    let oc = open_out_gen [Open_wronly; Open_creat] perms sess_file in
    output_string oc (md5(user_agent ^ ip) ^ "\n");
    close_out oc;
  with _ -> ()
;;

let create_sess_id() =
  let user_agent =
    try Unix.getenv "HTTP_USER_AGENT" with Not_found -> ""
  and ip =
    try Unix.getenv "REMOTE_ADDR" with Not_found -> ""
  in
  let rand_id =
    Random.self_init();
    string_of_int(Random.int max_int)
  in
  let sess_start = string_of_float start in
  let str = (user_agent ^ ip ^ rand_id ^ sess_start) in
  let sess_id = md5 str in
  make_sess_file ~sess_id ~user_agent ~ip;
  (sess_id)
;;

let sess_of_url ~incoming_args =
  let sess_id = List.assoc "sess" incoming_args in
  let sess_file = ("/tmp/sess_" ^ sess_id) in
  if not (Sys.file_exists sess_file)
  then raise Not_found
  else begin
    let user_agent =
      try Unix.getenv "HTTP_USER_AGENT" with Not_found -> ""
    and ip =
      try Unix.getenv "REMOTE_ADDR" with Not_found -> ""
    in
    let ic = open_in("/tmp/sess_" ^ sess_id) in
    let id = input_line ic in
    if id <> md5(user_agent ^ ip)
    then (close_in ic; raise Not_found)
    else begin
      let rec get_sess_env acc =
        try
          let l = input_line ic in
          get_sess_env(l::acc)
        with End_of_file -> acc
      in
      let sess_env =
        let raw = get_sess_env [] in
        parse_sess_env raw
      in
      close_in ic;
      (sess_id, sess_env)
    end
  end
;;

let (sess_id, sess_env) =
  let sess =
    try sess_of_url ~incoming_args;
    with _ -> create_sess_id(), []
  in
  (sess)
;;

(* {{{ DEBUG *(

let () =
  echo "<!-- SESS ENV\n";
  List.iter (fun (k,v) -> Printf.printf "  %s => '%s'\n" k v) sess_env;
  echo "-->\n";
;;
)* }}} *)

(* }}} *)
(* {{{ static pages *)

(* {{{ about *)

let about() =
  echo "\
<div class='speach'>
<h2>About</h2>
<p>This is the first alpha version of this site,
so it is currently very basic.<br />
The goal is both to register the maximum of free creations available on the web,
and to host free creations that authors want to share.</p>
</div>
"
;;
(* }}} *)
(* {{{ contribute *)

let contribute() =
  let datas_filename = datas_path ^ "datas.xml" in
  echo "\
<div class='speach'>
<h2>Contribute</h2>
<p>If you want to contribute in a way or another,
just send me an email:</p>
<div id='contact'>&nbsp;</div>

<p>Here are some exemple of way you can contribute:</p>
<ul>
<li>by telling me where Free creations can be found</li>
<li>by sending patch for the
	<a href='./free-creations.alpha.0.01.tgz'>back-end</a> /
	<a href='"; echo datas_filename; echo "'>DB</a></li>
<li>by sending a new CSS which will be included in a CSS switcher</li>
</ul>
</div>
"
;;
(* }}} *)
(* {{{ crawler *)

let crawler() =
  echo "\
<div class='speach'>
<h2>Milky Crawl</h2>
<p>
	The \"Milky&nbsp;Crawl\" web-crawler takes care of <em>robots.txt</em> files.<br />
	The user-agent string you should use in your robots.txt file to refer to it is
	<span class='highlight'>Milky&nbsp;Crawl</span>.
</p>
</div>
"
;;
(* }}} *)

(* }}} *)
(* {{{ preferences *)

(* {{{ preference form *)

let print_pref_form() =
  echos ["\
<div class='speach'>
<h2>Preferences</h2>

<form action='"; script_name; "' method='POST'>
	<input type='hidden' name='o' value='preferences' />
	<input type='hidden' name='sess' value='"; sess_id; "' />
	<fieldset>
		<legend accesskey='N'>Number of items per page</legend>
		<select name='var_td'>
			<option value=''>Colomns</option>
			<option value='2'>2 columns</option>
			<option value='3'>3 columns</option>
			<option value='4'>4 columns</option>
			<option value='5'>5 columns</option>
		</select>
		<select name='var_tr'>
			<option value=''>Rows</option>
			<option value='2'>2 rows</option>
			<option value='3'>3 rows</option>
			<option value='4'>4 rows</option>
			<option value='5'>5 rows</option>
			<option value='6'>6 rows</option>
			<option value='7'>7 rows</option>
			<option value='8'>8 rows</option>
		</select>
	</fieldset>
	<!--
	<fieldset>
		<legend accesskey='S'>Thumbnails maximized size</legend>
		<select name='var_tnsize'>
			<option value=''>width x height in pixels</option>
			<option value='200x180'>200 x 180 px</option>
			<option value='120x150'>120 x 150 px</option>
			<option value='480x420'>480 x 420 px</option>
		</select>
	</fieldset>
	-->
	<!--
	<fieldset>
		<legend accesskey='L'>Language</legend>
		<select name='var_lang'>
			<option value=''></option>
			<option value='lang_en'>english</option>
			<option value='lang_fr'>fran&ccedil;ais</option>
		</select>
	</fieldset>
	-->
	<!--
	<fieldset>
		<legend accesskey='C'>Graphic Style</legend>
		<select name='var_css'>
			<option value=''></option>
			<option value='01'>01</option>
			<option value='02'>02</option>
		</select>
	</fieldset>
	-->
	<input type='submit' />
</form>

</div>
"];
;;
(* }}} *)
(* {{{ set preferences *)

let sess_record ~var ~label =
  let label = String.uppercase label
  and sess_file = ("/tmp/sess_" ^ sess_id)
  and perms = 0o644
  in
  let oc = open_out_gen [Open_append] perms sess_file in
  output_string oc (label ^ "=" ^ var ^ "\n");
  close_out oc;
;;

let recieved = function
  | None -> (fun _ _ -> ())
  | Some var ->
      (fun label check_var ->
        try
          let var = check_var var in
          sess_record ~var ~label;
          echos ["<div>"; label; " = "; var; "</div>\n"];
        with _ -> ()
      )
;;
(* }}} *)

let get_form_var name =
  try
    let arg = List.assoc name incoming_args in
    if arg = "" then None else Some arg
  with Not_found -> None
;;

(* {{{ check user inputs *)

let check_int_range _min _max var =
  let var = int_of_string var in
  let var = min var _max in
  let var = max var _min in
  let var = string_of_int var in
  (var)
;;

let check_thumb_size = function
  | "120x150" -> "120x150"
  | "480x420" -> "480x420"
  | "200x180"
  | _ -> "200x180"
;;

let check_lang = function
  | "en" -> "en"
  | "fr" -> "fr"
  | _ -> "en"
;;

let check_css = function
  | "01" -> "01"
  | "02" -> "02"
  | _ -> "01"
;;

(* }}} *)

let preferences() =
  let pref_td   = get_form_var "var_td"
  and pref_tr   = get_form_var "var_tr"
  and pref_lang = get_form_var "var_lang"
  and pref_tns  = get_form_var "var_tnsize"
  and pref_css  = get_form_var "var_css"
  in
  match (pref_td, pref_tr, pref_lang, pref_tns, pref_css) with
  | None, None, None, None, None ->
      print_pref_form()
  | _ ->
      echo "<div class='speach'>\n\
            <h2>Preferences Set</h2>\n";

      recieved pref_lang "Language" (check_lang);
      recieved pref_css  "CSS"      (check_css);
      recieved pref_td   "Columns"  (check_int_range 1 5);
      recieved pref_tr   "Rows"     (check_int_range 1 8);
      recieved pref_tns  "Thumbnails size" (check_thumb_size);

      echo "</div>\n";
;;
(* }}} *)
(* {{{ index random *)

(* {{{ items number *)

let items_number() =

  (* load from index file *)
  let (idx_nb) =
    let fd =
      let index_filename = datas_path ^ "datas.idx" in
      (Unix.openfile index_filename [Unix.O_RDONLY] 0o640)
    in
    let size =
      let stats = Unix.fstat fd in
      (stats.Unix.st_size)
    in
    let item_size =
      let ic = Unix.in_channel_of_descr fd in
      let pos = input_byte ic in
      let len = input_byte ic in
      (pos + len)
    in
    let nb = (size - 2) / item_size in
    Unix.close fd;
    (nb)
  in
  echos["<!-- "; string_of_int idx_nb; " items in the database -->\n"];

  (idx_nb)
;;

(* }}} *)

let get_pref conv var_name default =
  try conv(List.assoc var_name sess_env)
  with Not_found -> default
;;

(* {{{ get one item *)

let get_item ~n =
  (* {{{ get the index & length of the item tag n *)

  let (idx, len) =
    let fd =
      let index_filename = datas_path ^ "datas.idx" in
      (Unix.openfile index_filename [Unix.O_RDONLY] 0o640)
    in
    let pos = 2 + (n * 5) in (* TODO: 5 from the 2 first octets *)

    let here = Unix.lseek fd pos Unix.SEEK_SET in
    if here <> pos then
      failwith "seek in index failed";

    let ic = Unix.in_channel_of_descr fd in
    let a = input_byte ic in
    let b = input_byte ic in
    let c = input_byte ic in
    let d = input_byte ic in
    let e = input_byte ic in
    let idx = a + (b lsl 8) + (c lsl 8)
    and len = d + (e lsl 8) in

    Unix.close fd;

    (idx, len)
  in
  (* }}} *)
  (* {{{ read the database file to get the item tag *)

  let fd =
    let index_filename = datas_path ^ "datas.xml" in
    (Unix.openfile index_filename [Unix.O_RDONLY] 0o640)
  in
  let here = Unix.lseek fd idx Unix.SEEK_SET in
  if here <> idx then
    failwith "seek in database failed";

  let buf = String.create len in
  let got = Unix.read fd buf 0 len in
  if got <> len then
    failwith "partial datas read";

  (* DEBUG:
  Printf.printf " <!-- GOT RAW:\n '%s'\n -->\n" buf;
  *)
  (* }}} *)
  (* {{{ parse the XML item tag *)

  let parse_item ~attrs =
    let attr name = try List.assoc name attrs with Not_found -> "" in
    let thumbnail = attr "thumbnail"
    and website   = attr "website"
    and location  = attr "location"
    and copy      = attr "copy"
    and licence   = attr "licence"
    in
    (thumbnail, website, location, copy, licence)
  in
  let item =
    let d = Xml.parse_string buf in
    match d with
    | Xml.Element(tag,attrs,_) ->
        if tag <> "item" then
          echos["<!-- warning: tag '"; tag; "' found -->\n"];
        let item = parse_item ~attrs in
        (item)
    | Xml.PCData str ->
        failwith("should be an item tag, but PCData found: " ^ str)
  in
  (* }}} *)
  Unix.close fd;
  (item)
;;

(* }}} *)
(* {{{ get a random set *)

let get_rand_set ~nb_items ~nb =
  Random.self_init();
  let rec loop acc i =
    if i >= nb then acc else
      (* take one not seen yet *)
      let rec search_not_seen() =
        let n = Random.int nb_items in
        if List.mem n acc then search_not_seen() else n
      in
      let n = search_not_seen() in
      loop (n::acc) (succ i)
  in
  loop [] 0
;;
(* }}} *)

let random () =
  let (nb_items) = items_number() in

  let thumb_path = "/LO/FC/thumbnails_200x180/" in

  let nb_td = get_pref int_of_string "COLUMNS" 4
  and nb_tr = get_pref int_of_string "ROWS" 3 in

  let rand_set = get_rand_set ~nb_items ~nb:(nb_td * nb_tr) in

  echo "<table id='tnset'>\n<tr>\n";

  (* {{{ print a table cell *)

  let print_cell ~todo =
    try
      let (n,todo) =
        (* {{{ take one not seen yet *)
        match todo with
        | hd::tl -> (hd,tl)
        | [] -> failwith "nothing to do"
        (* }}} *)
      in
      let (thumbnail, website, location, copy, licence) = get_item ~n in
      let licence_url = get_licence_url ~licence in
      echo  ("<td class='item'><div>\n");
      echos ["<a href='http://"; website; location; "' target='_blank'>"];
      echos ["<img src='"; thumb_path; thumbnail; "' class='thumb' /></a></div>\n"];
      echos ["<div class='legend'>\
              <span class='licence'><span class='"; licence; "'>\
              <a href='"; licence_url; "'>"; licence; "</a></span></span> \
              <span class='copy'>&copy; "; copy; "</span></div></td>\n\n"];
      (todo)
    with Failure("nothing to do") -> []
  in
  (* }}} *)

  let rec td_loop todo td =
    if td < nb_td then
      let todo = print_cell ~todo in
      td_loop (todo) (succ td)
    else
      todo
  in
  let rec tr_loop todo tr =
    if tr < nb_tr then
      let todo = td_loop todo 0
      and () = echo "</tr><tr>\n" in
      tr_loop todo (succ tr)
  in
  tr_loop rand_set 0;
  (* {{{ a more simple form *(

  for tr = 0 to pred nb_tr do
    for td = 0 to pred nb_td do
      ignore(print_cell []);
    done;
    echo "</tr><tr>\n";
  done;

  )* }}} *)

  echo "</tr>\n</table>\n";

  echos ["<div style='text-align:right;'>\
          <a href='"; script_name; "?sess="; sess_id; "' title='display another set'>\
          <span>Next set</span></a></div>"];
;;
(* }}} *)

(* {{{ html head *)

let () =
  let style_css =
    try List.assoc "CSS" sess_env
    with Not_found -> "01"
  in
  echos["\
<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">
<html xmlns='http://www.w3.org/1999/xhtml' xml:lang='en'>
<head>
	<meta http-equiv='content-type' content='text/html; charset=iso-8859-1' />
	<title>Free Creations</title>" ^
(*
"<base href='http://free-creations.blue-prawn.org/' />"
*)
"	<meta name='description' content='A web-site dedicated to free creations.' />
	<meta name='keywords' content='creation, free, photo, graphics' />
	<meta name='author' content='Florent Monnier' />
	<meta name='robots' content='all' />

	<style type='text/css' media='all'>
		@import \"/LO/FC/styles/"; style_css; "/index.css\";
	</style>
</head>
<body>
<div id='header'>
	<h1><a href='"; script_name; "?sess="; sess_id; "'><span>Free Creations</span></a></h1>
	<span id='version'>alpha</span>
</div>
<div id='menu'>
<ul>
	<li><a href='"; script_name; "?sess="; sess_id; "' title='display a random set'><span>Random browse</span></a></li>
	<li><a href='"; script_name; "?o=about&sess="; sess_id; "'><span>About</span></a></li>
	<li><a href='"; script_name; "?o=contribute&sess="; sess_id; "'><span>Contribute</span></a></li>
	<li><a href='"; script_name; "?o=preferences&sess="; sess_id; "'><span>Preferences</span></a></li>
</ul>
</div>
<div id='content'>
"];

  log_domain();

  (* }}} *)
  (* {{{ select the page content *)

  let page =
    try List.assoc "o" incoming_args
    with Not_found -> "index"
  in
  begin
    try
      match page with
      | "about" -> about()
      | "contribute" -> contribute()
      | "milky_crawl" -> crawler()
      | "preferences" -> preferences()
      | "index"
      | _ -> random()
    with e ->
      echos ["<!-- "; Printexc.to_string e; " -->\n"];
      about()
  end;

  (* }}} *)
  (* {{{ end of page *)
  echo "\
</div>
<div id='footer'>Creations &copy; respective owners.</div>
";
(* {{{ bench *)

  let finish = Unix.gettimeofday() in
  Printf.printf "<div id='bench'>exec-time = %.6fs</div>\n" (finish -. start);

(* }}} *)
  echo "\
</body>
</html>
";
(* }}} *)

(* vim: sw=2 sts=2 ts=2 et fdm=marker
 *)
